@@ -1,16 +1,19 @@
 import os
-import hjson
 import tempfile
+import warnings
+import hjson
+from typing import Callable, Dict, Any
 import tensorflow as tf
 import numpy as np
 from c3.signal.pulse import Envelope, Carrier
 from c3.signal.gates import Instruction
-from c3.c3objs import Quantity, C3obj
+from c3.c3objs import Quantity, C3obj, hjson_encode
+from c3.utils.tf_utils import tf_convolve
 
 devices = dict()
 
 
-def dev_reg_deco(func):
+def dev_reg_deco(func: Callable) -> Callable:
     """
     Decorator for making registry of functions
     """
@@ -36,7 +39,11 @@ class Device(C3obj):
         name = props.pop("name")
         desc = props.pop("desc", "")
         comment = props.pop("comment", "")
-        params = props
+
+        # Because of legacy usage, we might have parameters given withing props iself
+        # or in a "params" field. Here we combine them.
+        params = props.pop("params", {})
+        params.update(props)
         super().__init__(name, desc, comment, params)
         self.signal = {}
 
@@ -45,9 +52,9 @@ class Device(C3obj):
         Write dictionary to a HJSON file.
         """
         with open(filepath, "w") as cfg_file:
-            hjson.dump(self.asdict(), cfg_file)
+            hjson.dump(self.asdict(), cfg_file, default=hjson_encode)
 
-    def asdict(self) -> dict:
+    def asdict(self) -> Dict[str, Any]:
         params = {}
         for key, item in self.params.items():
             params[key] = item.asdict()
@@ -60,9 +67,9 @@ class Device(C3obj):
         }
 
     def __str__(self) -> str:
-        return hjson.dumps(self.asdict())
+        return hjson.dumps(self.asdict(), default=hjson_encode)
 
-    def calc_slice_num(self, t_start: np.float64, t_end: np.float64):
+    def calc_slice_num(self, t_start: np.float64, t_end: np.float64) -> None:
         """
         Effective number of time slices given start, end and resolution.
 
@@ -77,7 +84,9 @@ class Device(C3obj):
         self.slice_num = int(np.abs(t_start - t_end) * res)
         # return self.slice_num
 
-    def create_ts(self, t_start: np.float64, t_end: np.float64, centered: bool = True):
+    def create_ts(
+        self, t_start: np.float64, t_end: np.float64, centered: bool = True
+    ) -> tf.constant:
         """
         Compute time samples.
 
@@ -90,8 +99,9 @@ class Device(C3obj):
         centered: boolean
             Sample in the middle of an interval, otherwise at the beginning.
         """
-        if not hasattr(self, "slice_num"):
-            self.calc_slice_num(t_start, t_end)
+
+        # Slice num can change between pulses
+        self.calc_slice_num(t_start, t_end)
         dt = 1 / self.resolution
         # TODO This type of centering does not guarantee zeros at the ends
         if centered:
@@ -100,8 +110,16 @@ class Device(C3obj):
         else:
             offset = 0
             num = self.slice_num + 1
-        t_start = tf.Variable(t_start + offset, dtype=tf.float64)
-        t_end = tf.Variable(t_end - offset, dtype=tf.float64)
+        t_start = tf.constant(t_start + offset, dtype=tf.float64)
+        t_end = tf.constant(t_end - offset, dtype=tf.float64)
+        np.testing.assert_almost_equal(
+            np.mod(t_end, dt),
+            0,
+            decimal=7,
+            err_msg="Given length of is not a multiple of the resolution",
+        )
+
+        # ts = tf.range(t_start, t_end + 1e-16, dt)
         ts = tf.linspace(t_start, t_end, num)
         return ts
 
@@ -159,7 +177,9 @@ class VoltsToHertz(Device):
         self.inputs = props.pop("inputs", 1)
         self.outputs = props.pop("outputs", 1)
 
-    def process(self, instr, chan, mixed_signal):
+    def process(
+        self, instr: Instruction, chan: str, mixed_signal: Dict[str, Any]
+    ) -> Dict[str, Any]:
         """Transform signal from value of V to Hz.
 
         Parameters
@@ -178,6 +198,75 @@ class VoltsToHertz(Device):
         return self.signal
 
 
+@dev_reg_deco
+class Crosstalk(Device):
+    """
+    Device to phenomenologically include crosstalk in the model by explicitly mixing
+    drive lines.
+
+    Parameters
+    ----------
+
+    crosstalk_matrix: tf.constant
+        Matrix description of how to mix drive channels.
+
+    Examples
+    --------
+    .. code-block:: python
+
+        xtalk = Crosstalk(
+            name="crosstalk",
+            channels=["TC1", "TC2"],
+            crosstalk_matrix=Quantity(
+                value=[[1, 0], [0, 1]],
+                min_val=[[0, 0], [0, 0]],
+                max_val=[[1, 1], [1, 1]],
+                unit="",
+            ),
+        )
+
+
+
+    """
+
+    def __init__(self, **props):
+        self.crossed_channels = props.pop("channels", None)
+        super().__init__(**props)
+        self.inputs = props.pop("inputs", 1)
+        self.outputs = props.pop("outputs", 1)
+        self.params["crosstalk_matrix"] = props.pop("crosstalk_matrix", None)
+
+    def process(self, signal: Dict[str, Any]) -> Dict[str, Any]:
+        """
+        Mix channels in the input signal according to a crosstalk matrix.
+
+        Parameters
+        ----------
+        signal: Dict[str, Any]
+            Dictionary of several signals identified by their channel as dict keys, e.g.
+
+            .. code-block:: python
+
+                signal = {
+                    "TC1": {"values": [0, 0.5, 1, 1, ...]},
+                    "TC2": {"values": [1, 1, 1, 1, ...],
+                }
+
+
+
+        Returns
+        -------
+        signal: Dict[str, Any]
+
+        """
+        xtalk = self.params["crosstalk_matrix"]
+        signals = [signal[ch]["values"] for ch in self.crossed_channels]
+        crossed_signals = xtalk.get_value() @ signals
+        for indx, ch in enumerate(self.crossed_channels):
+            signal[ch]["values"] = crossed_signals[indx]
+        return signal
+
+
 @dev_reg_deco
 class DigitalToAnalog(Device):
     """Take the values at the awg resolution to the simulation resolution."""
@@ -189,17 +278,20 @@ class DigitalToAnalog(Device):
         self.ts = None
         self.sampling_method = props.pop("sampling_method", "nearest")
 
-    def process(self, instr, chan, awg_signal):
+    def process(
+        self, instr: Instruction, chan: str, awg_signal: Dict[str, Any]
+    ) -> Dict[str, Any]:
         """Resample the awg values to higher resolution.
 
         Parameters
         ----------
-        awg_signal: tf.Tensor
-            Bandwith-limited, low-resolution AWG signal.
-        t_start: np.float64
-            Beginning of the signal.
-        t_end: np.float64
-            End of the signal.
+        instr: Instruction
+            The logical instruction or qubit operation for which the signal is
+            generated.
+        chan: str
+            Specifies which channel is being processed if needed.
+        awg_signal: dict
+            Dictionary of several signals identified by their channel as dict keys.
 
         Returns
         -------
@@ -218,7 +310,7 @@ class DigitalToAnalog(Device):
             ),
             shape=[new_dim],
         )
-        inphase = tf.cast(inphase, dtype=tf.float64)
+        inphase = tf.cast(inphase, tf.float64)
         quadrature = tf.reshape(
             tf.image.resize(
                 tf.reshape(awg_signal["quadrature"], shape=[1, old_dim, 1]),
@@ -227,7 +319,7 @@ class DigitalToAnalog(Device):
             ),
             shape=[new_dim],
         )
-        quadrature = tf.cast(quadrature, dtype=tf.float64)
+        quadrature = tf.cast(quadrature, tf.float64)
         self.signal["ts"] = ts
         self.signal["inphase"] = inphase
         self.signal["quadrature"] = quadrature
@@ -236,15 +328,17 @@ class DigitalToAnalog(Device):
 
 @dev_reg_deco
 class Filter(Device):
-    # TODO This can apply a general function to a signal.
+    # TODO This can apply a general function to a signal. --> Should merge into StepFuncFilter
     """Apply a filter function to the signal."""
 
     def __init__(self, **props):
         raise Exception("C3:ERROR Not yet implemented.")
-        self.filter_function = props["filter_function"]
+        self.filter_function: Callable = props["filter_function"]
         super().__init__(**props)
 
-    def process(self, instr, chan, Hz_signal):
+    def process(
+        self, instr: Instruction, chan: str, Hz_signal: Dict[str, Any]
+    ) -> Dict[str, Any]:
         """Apply a filter function to the signal."""
         self.signal = self.filter_function(Hz_signal)
         return self.signal
@@ -270,12 +364,35 @@ class FluxTuning(Device):
         super().__init__(**props)
         self.inputs = props.pop("inputs", 1)
         self.outputs = props.pop("outputs", 1)
-        for par in ["phi_0", "phi", "omega_0"]:
+        for par in ["phi_0", "phi", "omega_0", "anhar"]:
             if par not in self.params:
                 raise Exception(
                     f"C3:ERROR: {self.__class__}  needs a '{par}' parameter."
                 )
 
+    def get_factor(self, phi):
+        pi = tf.constant(np.pi, dtype=tf.float64)
+        phi_0 = tf.cast(self.params["phi_0"].get_value(), tf.float64)
+
+        if "d" in self.params:
+            d = self.params["d"].get_value()
+            factor = tf.sqrt(
+                tf.sqrt(
+                    tf.cos(pi * phi / phi_0) ** 2
+                    + d ** 2 * tf.sin(pi * phi / phi_0) ** 2
+                )
+            )
+        else:
+            factor = tf.sqrt(tf.abs(tf.cos(pi * phi / phi_0)))
+        return factor
+
+    def get_freq(self, phi):
+        # TODO: Check how the time dependency affects the frequency. (Koch et al. , 2007)
+        omega_0 = self.params["omega_0"].get_value()
+        anhar = self.params["anhar"].get_value()
+        biased_freq = (omega_0 - anhar) * self.get_factor(phi) + anhar
+        return biased_freq
+
     def process(self, instr: Instruction, chan: str, signal_in):
         """
         Compute the qubit frequency resulting from an applied flux.
@@ -290,38 +407,10 @@ class FluxTuning(Device):
         tf.float64
             Qubit frequency.
         """
-        pi = tf.constant(np.pi, dtype=tf.float64)
         phi = self.params["phi"].get_value()
-        omega_0 = self.params["omega_0"].get_value()
-        phi_0 = self.params["phi_0"].get_value()
         signal = signal_in["values"]
         self.signal["ts"] = signal_in["ts"]
-
-        if "d" in self.params:
-            d = self.params["d"].get_value()
-            #             print('assuming asymmetric transmon with d=', d)
-            base_freq = omega_0 * tf.sqrt(
-                tf.sqrt(
-                    tf.cos(pi * phi / phi_0) ** 2
-                    + d ** 2 * tf.sin(pi * phi / phi_0) ** 2
-                )
-            )
-            freq = (
-                omega_0
-                * tf.sqrt(
-                    tf.sqrt(
-                        tf.cos(pi * (phi + signal) / phi_0) ** 2
-                        + d ** 2 * tf.sin(pi * (phi + signal) / phi_0) ** 2
-                    )
-                )
-                - base_freq
-            )
-        else:
-            base_freq = omega_0 * tf.sqrt(tf.abs(tf.cos(pi * phi / phi_0)))
-            freq = (
-                omega_0 * tf.sqrt(tf.abs(tf.cos(pi * (phi + signal) / phi_0)))
-                - base_freq
-            )
+        freq = self.get_freq(phi + signal) - self.get_freq(phi)
         self.signal["values"] = freq
         return self.signal
 
@@ -350,7 +439,7 @@ class FluxTuningLinear(Device):
                 )
         self.freq = None
 
-    def frequency(self, signal):
+    def frequency(self, signal: tf.float64) -> tf.constant:
         """
         Compute the qubit frequency resulting from an applied flux.
 
@@ -381,6 +470,7 @@ class FluxTuningLinear(Device):
         return self.freq
 
 
+# Depreciated use ResponseFFT instead
 @dev_reg_deco
 class Response(Device):
     """Make the AWG signal physical by convolution with a Gaussian to limit bandwith.
@@ -395,6 +485,9 @@ class Response(Device):
         super().__init__(**props)
         self.inputs = props.pop("inputs", 1)
         self.outputs = props.pop("outputs", 1)
+        warnings.warn(
+            "use ResponseFFT for faster signal generation instead", DeprecationWarning
+        )
 
     def convolve(self, signal: list, resp_shape: list):
         """
@@ -456,7 +549,7 @@ class Response(Device):
         """
         n_ts = tf.floor(self.params["rise_time"].get_value() * self.resolution)
         ts = tf.linspace(
-            tf.Variable(0.0, dtype=tf.float64),
+            tf.constant(0.0, dtype=tf.float64),
             self.params["rise_time"].get_value(),
             tf.cast(n_ts, tf.int32),
         )
@@ -480,6 +573,155 @@ class Response(Device):
         return self.signal
 
 
+@dev_reg_deco
+class ResponseFFT(Device):
+    """Make the AWG signal physical by convolution with a Gaussian to limit bandwith.
+
+    Parameters
+    ----------
+    rise_time : Quantity
+        Time constant for the gaussian convolution.
+    """
+
+    def __init__(self, **props):
+        super().__init__(**props)
+        self.inputs = props.pop("inputs", 1)
+        self.outputs = props.pop("outputs", 1)
+
+    def process(self, instr, chan, iq_signal):
+        """
+        Apply a Gaussian shaped limiting function to an IQ signal.
+
+        Parameters
+        ----------
+        iq_signal : dict
+            I and Q components of an AWG signal.
+
+        Returns
+        -------
+        dict
+            Bandwidth limited IQ signal.
+
+        """
+        np.testing.assert_almost_equal(
+            actual=iq_signal["ts"][1] - iq_signal["ts"][0], desired=1 / self.resolution
+        )
+        n_ts = tf.floor(self.params["rise_time"].get_value() * self.resolution)
+        ts = tf.linspace(
+            tf.constant(0.0, dtype=tf.float64),
+            self.params["rise_time"].get_value(),
+            tf.cast(n_ts, tf.int32),
+        )
+        cen = tf.cast(
+            (self.params["rise_time"].get_value() - 1 / self.resolution) / 2, tf.float64
+        )
+        sigma = self.params["rise_time"].get_value() / 4
+        gauss = tf.exp(-((ts - cen) ** 2) / (2 * sigma * sigma))
+        offset = tf.exp(-((-1 - cen) ** 2) / (2 * sigma * sigma))
+
+        risefun = gauss - offset
+        inphase = tf_convolve(iq_signal["inphase"], risefun / tf.reduce_sum(risefun))
+        quadrature = tf_convolve(
+            iq_signal["quadrature"], risefun / tf.reduce_sum(risefun)
+        )
+
+        inphase = tf.math.real(inphase)
+        quadrature = tf.math.real(quadrature)
+        self.signal = {
+            "inphase": inphase,
+            "quadrature": quadrature,
+            "ts": iq_signal["ts"],
+        }
+        return self.signal
+
+
+@dev_reg_deco
+class StepFuncFilter(Device):
+    """
+    Base class for filters that are based on the step response function
+    Step function has to be defined explicetly
+    """
+
+    def __init__(self, **props):
+        super().__init__(**props)
+        self.inputs = props.pop("inputs", 1)
+        self.outputs = props.pop("outputs", 1)
+
+    def step_response_function(self, ts):
+        raise NotImplementedError()
+
+    def process(self, instr, chan, signal_in):
+        ts = tf.identity(signal_in["ts"])
+        step_response = self.step_response_function(ts)
+        step_response = tf.concat([[0], step_response], axis=0)
+        impulse_response = step_response[1:] - step_response[:-1]
+        signal_out = dict()
+        for key, signal in signal_in.items():
+            if key == "ts":
+                continue
+            signal_out[key] = tf.cast(tf_convolve(signal, impulse_response), tf.float64)
+        signal_out["ts"] = signal_in["ts"]
+
+        return signal_out
+
+
+@dev_reg_deco
+class ExponentialIIR(StepFuncFilter):
+    """Implement IIR filter with step response of the form
+    s(t) = (1 + A * exp(-t / t_iir) )
+
+    Parameters
+    ----------
+    time_iir: Quantity
+        Time constant for the filtering.
+    amp: Quantity
+
+    """
+
+    def step_response_function(self, ts):
+        time_iir = self.params["time_iir"]
+        amp = self.params["amp"]
+        step_response = 1 + amp * tf.exp(-ts / time_iir)
+        return step_response
+
+
+@dev_reg_deco
+class HighpassExponential(StepFuncFilter):
+    """Implement Highpass filter based on exponential with step response of the form
+    s(t) = exp(-t / t_hp)
+
+    Parameters
+    ----------
+    time_iir: Quantity
+        Time constant for the filtering.
+    amp: Quantity
+
+    """
+
+    def step_response_function(self, ts):
+        time_hp = self.params["time_hp"]
+        return tf.exp(-ts / time_hp)
+
+
+@dev_reg_deco
+class SkinEffectResponse(StepFuncFilter):
+    """Implement Highpass filter based on exponential with step response of the form
+    s(t) = exp(-t / t_hp)
+
+    Parameters
+    ----------
+    time_iir: Quantity
+        Time constant for the filtering.
+    amp: Quantity
+
+    """
+
+    def step_response_function(self, ts):
+        alpha = self.params["alpha"]
+        return tf.math.erfc(alpha / 21 / tf.math.sqrt(np.abs(ts)))
+
+
+# Obsolete. Use HighpassExponential
 class HighpassFilter(Device):
     """Introduce a highpass filter
 
@@ -540,7 +782,7 @@ class HighpassFilter(Device):
             )
         return convolution
 
-    def process(self,  instr, chan, iq_signal):
+    def process(self, instr, chan, iq_signal):
         """
         Apply a highpass cutoff to an IQ signal.
 
@@ -565,27 +807,31 @@ class HighpassFilter(Device):
         # fc = 1e7 / self.resolution
         # tb = fc / 2
 
-        N_ts = tf.cast(tf.math.ceil(4 / tb), dtype=tf.int32)
+        N_ts = tf.cast(tf.math.ceil(4 / tb), tf.int32)
         N_ts += 1 - tf.math.mod(N_ts, 2)  # make n_ts odd
         if N_ts > len(iq_signal["inphase"] * 100):
             self.signal = iq_signal
             return self.signal
 
-        pi = tf.cast(np.pi, dtype=tf.double)
+        pi = tf.cast(np.pi, tf.double)
 
-        n = tf.cast(tf.range(N_ts), dtype=tf.double)
+        n = tf.cast(tf.range(N_ts), tf.double)
 
         x = 2 * fc * (n - (N_ts - 1) / 2)
         h = tf.sin(pi * x) / (pi * x)
         h = tf.where(tf.math.is_nan(h), tf.ones_like(h), h)
         w = tf.signal.hamming_window(N_ts)
-        w = tf.cast(w, dtype=tf.double)
+        w = tf.cast(w, tf.double)
         h *= w
         h /= -tf.reduce_sum(h)
-        h = tf.where(tf.cast(n, dtype=tf.int32) == (N_ts - 1) // 2, tf.ones_like(h), h)
+        h = tf.where(tf.cast(n, tf.int32) == (N_ts - 1) // 2, tf.ones_like(h), h)
         inphase = self.convolve(iq_signal["inphase"], h)
         quadrature = self.convolve(iq_signal["quadrature"], h)
-        self.signal = {"inphase": inphase, "quadrature": quadrature, "ts": iq_signal["ts"]}
+        self.signal = {
+            "inphase": inphase,
+            "quadrature": quadrature,
+            "ts": iq_signal["ts"],
+        }
         return self.signal
 
 
@@ -634,7 +880,7 @@ class LONoise(Device):
         self.signal = None
         self.params["noise_perc"] = props.pop("noise_perc")
 
-    def process(self,  instr, chan, lo_signal):
+    def process(self, instr, chan, lo_signal):
         """Distort signal by adding noise."""
         noise_perc = self.params["noise_perc"].get_value()
         cos, sin = lo_signal["values"]
@@ -654,19 +900,26 @@ class Additive_Noise(Device):
         self.inputs = props.pop("inputs", 1)
         self.outputs = props.pop("outputs", 1)
         self.signal = None
-        self.params["noise_amp"] = props.pop("noise_amp")
+        if "noise_amp" in props:
+            self.params["noise_amp"] = props.pop("noise_amp")
 
-    def process(self,  instr, chan, signal):
+    def get_noise(self, sig):
+        noise_amp = self.params["noise_amp"].get_value()
+        return noise_amp * np.random.normal(size=tf.shape(sig), loc=0.0, scale=1.0)
+
+    def process(self, instr, chan, signal):
         """Distort signal by adding noise."""
         noise_amp = self.params["noise_amp"].get_value()
         out_signal = {"ts": signal["ts"]}
         for k, sig in signal.items():
-            if k != 'ts':
+            if k != "ts" and "noise" not in k:
                 if noise_amp < 1e-17:
                     noise = tf.zeros_like(sig)
                 else:
-                    noise = tf.constant(noise_amp * np.random.normal(size=tf.shape(sig), loc=0.0, scale=1.0))
-                noise_key = 'noise' + ('-' + k if k != "values" else "")
+                    noise = tf.constant(
+                        self.get_noise(sig), shape=sig.shape, dtype=tf.float64
+                    )
+                noise_key = "noise" + ("-" + k if k != "values" else "")
                 out_signal[noise_key] = noise
 
                 out_signal[k] = sig + noise
@@ -675,31 +928,41 @@ class Additive_Noise(Device):
 
 
 @dev_reg_deco
-class DC_Noise(Device):
-    """Noise applied to a signal"""
+class DC_Noise(Additive_Noise):
+    """Add a random constant offset to the signals"""
+
+    def get_noise(self, sig):
+        noise_amp = self.params["noise_amp"].get_value()
+        return tf.ones_like(sig) * tf.constant(
+            noise_amp * np.random.normal(loc=0.0, scale=1.0)
+        )
+
+
+@dev_reg_deco
+class Pink_Noise(Additive_Noise):
+    """Device creating pink noise, i.e. 1/f noise."""
 
     def __init__(self, **props):
         super().__init__(**props)
-        self.inputs = props.pop("inputs", 1)
-        self.outputs = props.pop("outputs", 1)
-        self.signal = None
-        self.params["noise_amp"] = props.pop("noise_amp", 0)
+        self.params["bfl_num"] = props.pop(
+            "bfl_num", Quantity(value=5, min_val=1, max_val=10)
+        )
 
-    def process(self,  instr, chan, signal):
-        """Distort signal by adding noise."""
-        noise_amp = self.params["noise_amp"].get_value()
-        out_signal = {"ts": signal["ts"]}
-        for k, sig in signal.items():
-            if k != 'ts':
-                if noise_amp < 1e-17:
-                    noise = tf.zeros_like(sig)
-                else:
-                    noise = tf.ones_like(sig) * tf.constant(noise_amp * np.random.normal(loc=0.0, scale=1.0))
-                noise_key = 'noise' + ('-' + k if k != "values" else "")
-                out_signal[noise_key] = noise
-                out_signal[k] = sig + noise
-        self.signal = out_signal
-        return self.signal
+    def get_noise(self, sig):
+        noise_amp = self.params["noise_amp"].get_value().numpy()
+        bfl_num = np.int(self.params["bfl_num"].get_value().numpy())
+        noise = []
+        bfls = 2 * np.random.randint(2, size=bfl_num) - 1
+        num_steps = len(sig)
+        flip_rates = np.logspace(
+            0, np.log(num_steps), num=bfl_num + 1, endpoint=True, base=10.0
+        )
+        for step in range(num_steps):
+            for indx in range(bfl_num):
+                if np.floor(np.random.random() * flip_rates[indx + 1]) == 0:
+                    bfls[indx] = -bfls[indx]
+            noise.append(np.sum(bfls) * noise_amp)
+        return noise
 
 
 @dev_reg_deco
@@ -713,7 +976,7 @@ class DC_Offset(Device):
         self.signal = None
         self.params["offset_amp"] = props.pop("offset_amp")
 
-    def process(self,  instr, chan, signal):
+    def process(self, instr, chan, signal):
         """Distort signal by adding noise."""
         offset_amp = self.params["offset_amp"].get_value()
         if np.abs(offset_amp) < 1e-17:
@@ -729,93 +992,6 @@ class DC_Offset(Device):
         return self.signal
 
 
-# TODO: We should write out own function to calculate the Pink noise in a continuous fft fashion.
-# import colorednoise
-# @dev_reg_deco
-# class Pink_Noise_Cont(Device):
-#     """Noise applied to a signal"""
-#
-#     def __init__(
-#             self,
-#             name: str = "pink_noise",
-#             desc: str = " ",
-#             comment: str = " ",
-#             resolution: np.float64 = 0.0,
-#             noise_amp: Quantity = None
-#     ):
-#         super().__init__(
-#             name=name,
-#             desc=desc,
-#             comment=comment,
-#             resolution=resolution
-#         )
-#         self.signal = None
-#         self.params['noise_amp'] = noise_amp
-#
-#     def distort(self, signal):
-#         """Distort signal by adding noise."""
-#         noise_amp = self.params['noise_amp'].get_value()
-#         if noise_amp < 1e-17:
-#             self.signal = signal
-#             return signal
-#         out_signal = {}
-#         # print(signal)
-#         if type(signal) is dict:
-#             for k, sig in signal.items():
-#                 out_signal[k] = sig + tf.constant(noise_amp * colorednoise.powerlaw_psd_gaussian(1,))
-#         else:
-#             out_signal = signal + tf.constant(noise_amp * np.random.normal(size=tf.shape(signal), loc=0.0, scale=1.0))
-#         self.signal = out_signal
-#         return self.signal
-
-
-@dev_reg_deco
-class Pink_Noise(Device):
-    """Device creating pink noise, i.e. 1/f noise."""
-
-    def __init__(self, **props):
-        super().__init__(**props)
-        self.inputs = props.pop("inputs", 1)
-        self.outputs = props.pop("outputs", 1)
-        self.signal = None
-        self.params["noise_strength"] = props.pop("noise_strength")
-        self.params["bfl_num"] = props.pop("bfl_num", Quantity(value=5, min_val=1, max_val=10))
-        self.ts = None
-        self.signal = None
-
-    def get_noise(self, sig, noise_strength, bfl_num):
-        noise = []
-        bfls = 2 * np.random.randint(2, size=bfl_num) - 1
-        num_steps = len(sig)
-        flip_rates = np.logspace(
-            0, np.log(num_steps), num=bfl_num + 1, endpoint=True, base=10.0
-        )
-        for step in range(num_steps):
-            for indx in range(bfl_num):
-                if np.floor(np.random.random() * flip_rates[indx + 1]) == 0:
-                    bfls[indx] = -bfls[indx]
-            noise.append(np.sum(bfls) * noise_strength)
-        return noise
-
-
-    def process(self, intr, chan, signal):
-        noise_strength = self.params["noise_strength"].get_value().numpy()
-        bfl_num = np.int(self.params["bfl_num"].get_value().numpy())
-
-        out_signal = {"ts": signal["ts"]}
-        for k, sig in signal.items():
-            if k != 'ts':
-                if noise_strength < 1e-17:
-                    noise = tf.zeros_like(sig)
-                else:
-                    noise = tf.constant(self.get_noise(sig, noise_strength, bfl_num), shape=sig.shape, dtype=tf.float64)
-                noise_key = 'noise' + ('-' + k if k != "values" else "")
-                out_signal[noise_key] = noise
-
-                out_signal[k] = sig + noise
-        self.signal = out_signal
-        return self.signal
-
 @dev_reg_deco
 class LO(Device):
     """Local oscillator device, generates a constant oscillating signal."""
@@ -842,7 +1018,7 @@ class LO(Device):
                 if amp_noise and freq_noise:
                     print("amp and freq noise")
                     phi = omega_lo * ts[0]
-                    for t in ts:
+                    for _ in ts:
                         A = np.random.normal(loc=1.0, scale=amp_noise)
                         cos.append(A * np.cos(phi))
                         sin.append(A * np.sin(phi))
@@ -869,7 +1045,7 @@ class LO(Device):
                         sin.append(np.sin(omega_lo * t + phi))
                 elif freq_noise:
                     phi = omega_lo * ts[0]
-                    for t in ts:
+                    for _ in ts:
                         cos.append(np.cos(phi))
                         sin.append(np.sin(phi))
                         omega = omega_lo + np.random.normal(loc=0.0, scale=freq_noise)
@@ -880,6 +1056,7 @@ class LO(Device):
                 self.signal["inphase"] = cos
                 self.signal["quadrature"] = sin
                 self.signal["ts"] = ts
+        assert "inphase" in self.signal, f"Probably no carrier proviced for {self.name}"
         return self.signal
 
 
@@ -946,108 +1123,17 @@ class AWG(Device):
 
         """
         ts = self.create_ts(instr.t_start, instr.t_end, centered=True)
-        components = instr.comps
         self.ts = ts
-        # dt = ts[1] - ts[0]
-        # t_before = ts[0] - dt
-        amp_tot_sq = 0.0
-        inphase_comps = []
-        quadrature_comps = []
 
-        for comp in components[chan].values():
-            if isinstance(comp, Envelope):
-
-                amp = comp.params["amp"].get_value()
-
-                amp_tot_sq += amp ** 2
-
-                xy_angle = comp.params["xy_angle"].get_value()
-                freq_offset = comp.params["freq_offset"].get_value()
-                phase = -xy_angle + freq_offset * ts
-                env = comp.get_shape_values(ts)
-                # TODO option to have t_before
-                # env = comp.get_shape_values(ts, t_before)
-                inphase_comps.append(amp * env * tf.cos(phase))
-                quadrature_comps.append(-amp * env * tf.sin(phase))
-
-        norm = tf.sqrt(tf.cast(amp_tot_sq, tf.float64))
-        inphase = tf.add_n(inphase_comps, name="inphase")
-        quadrature = tf.add_n(quadrature_comps, name="quadrature")
+        signal, norm = instr.get_awg_signal(chan, ts, options={self.__options: True})
 
         self.amp_tot = norm
-        self.signal[chan] = {"inphase": inphase, "quadrature": quadrature, "ts": ts}
-        return {"inphase": inphase, "quadrature": quadrature, "ts": ts}
-
-    def create_IQ_drag(self, instr: Instruction, chan: str) -> dict:
-        """
-        Construct the in-phase (I) and quadrature (Q) components of the signal.
-        These are universal to either experiment or simulation.
-        In the xperiment these will be routed to AWG and mixer
-        electronics, while in the simulation they provide the shapes of the
-        instruction fields to be added to the Hamiltonian.
-
-        Parameters
-        ----------
-        channel : str
-            Identifier for the selected drive line.
-        components : dict
-            Separate signals to be combined onto this drive line.
-        t_start : float
-            Beginning of the signal.
-        t_end : float
-            End of the signal.
-
-        Returns
-        -------
-        dict
-            Waveforms as I and Q components.
-
-        """
-        ts = self.create_ts(instr.t_start, instr.t_end, centered=True)
-        components = instr.comps
-        self.ts = ts
-        dt = ts[1] - ts[0]
-        t_before = ts[0] - dt
-        amp_tot_sq = 0.0
-        inphase_comps = []
-        quadrature_comps = []
-
-        for comp in components[chan].values():
-            if isinstance(comp, Envelope):
-
-                amp = comp.params["amp"].get_value()
-                amp_tot_sq += amp ** 2
-
-                xy_angle = comp.params["xy_angle"].get_value()
-                freq_offset = comp.params["freq_offset"].get_value()
-                # TODO should we remove this redefinition?
-                delta = -comp.params["delta"].get_value()
-                if self.__options == "drag_2":
-                    delta = delta * dt
-
-                with tf.GradientTape() as t:
-                    t.watch(ts)
-                    env = comp.get_shape_values(ts, t_before)
-                    # TODO option to have t_before = 0
-                    # env = comp.get_shape_values(ts, t_before)
-
-                denv = t.gradient(env, ts)
-                if denv is None:
-                    denv = tf.zeros_like(ts, dtype=tf.float64)
-                phase = -xy_angle + freq_offset * ts
-                inphase_comps.append(
-                    amp * (env * tf.cos(phase) + denv * delta * tf.sin(phase))
-                )
-                quadrature_comps.append(
-                    amp * (denv * delta * tf.cos(phase) - env * tf.sin(phase))
-                )
-        norm = tf.sqrt(tf.cast(amp_tot_sq, tf.float64))
-        inphase = tf.add_n(inphase_comps, name="inphase")
-        quadrature = tf.add_n(quadrature_comps, name="quadrature")
-
-        self.amp_tot = norm
-        self.signal[chan] = {"inphase": inphase, "quadrature": quadrature, "ts": ts}
-        return {"inphase": inphase, "quadrature": quadrature, "ts": ts}
+        self.signal[chan] = {
+            "inphase": signal["inphase"],
+            "quadrature": signal["quadrature"],
+            "ts": ts,
+        }
+        return self.signal[chan]
 
     def create_IQ_pwc(self, instr: Instruction, chan: str) -> dict:
         """
@@ -1082,7 +1168,6 @@ class AWG(Device):
         inphase_comps = []
         quadrature_comps = []
 
-        amp_tot_sq = 0
         for comp in components[chan].values():
             if isinstance(comp, Envelope):
                 amp_tot_sq += 1
@@ -1095,7 +1180,7 @@ class AWG(Device):
                     quadrature = tf.math.imag(shape)
                 xy_angle = comp.params["xy_angle"].get_value()
                 freq_offset = comp.params["freq_offset"].get_value()
-                phase = -xy_angle + freq_offset * ts
+                phase = xy_angle + freq_offset * ts
 
                 if len(inphase) != len(quadrature):
                     raise ValueError("inphase and quadrature are of different lengths.")
@@ -1144,14 +1229,13 @@ class AWG(Device):
         return self.signal[line]["quadrature"]  # * self.amp_tot
 
     def enable_drag(self):
-        self.process = self.create_IQ_drag
+        self.__options = "drag"
 
     def enable_drag_2(self):
-        self.process = self.create_IQ_drag
         self.__options = "drag_2"
 
     def enable_pwc(self):
-        self.process = self.create_IQ_pwc
+        self.__options = "pwc"
 
     def log_shapes(self):
         # TODO log shapes in the generator instead
@@ -1159,6 +1243,6 @@ class AWG(Device):
             signal = {}
             for key in self.signal:
                 signal[key] = self.signal[key].numpy().tolist()
-            logfile.write(hjson.dumps(signal))
+            logfile.write(hjson.dumps(signal, default=hjson_encode))
             logfile.write("\n")
             logfile.flush()
