@@ -1,11 +1,12 @@
 """Library of fidelity functions."""
-# TODO think of how to use the fidelity functions in a cleaner way
 
 import numpy as np
 import tensorflow as tf
+from typing import List, Dict
 
-# import matplotlib.pyplot as plt
 from scipy.optimize import curve_fit
+
+from c3.signal.gates import Instruction
 from c3.utils.tf_utils import (
     tf_ave,
     tf_super,
@@ -13,15 +14,19 @@ from c3.utils.tf_utils import (
     tf_ketket_fid,
     tf_superoper_unitary_overlap,
     tf_unitary_overlap,
+    tf_project_to_comp,
     tf_dm_to_vec,
     tf_average_fidelity,
     tf_superoper_average_fidelity,
     tf_state_to_dm,
+)
+
+from c3.libraries.propagation import (
     evaluate_sequences,
 )
+
 from c3.utils.qt_utils import (
     basis,
-    perfect_gate,
     perfect_cliffords,
     cliffords_decomp,
     cliffords_decomp_xId,
@@ -41,13 +46,15 @@ def fid_reg_deco(func):
 
 
 @fid_reg_deco
-def state_transfer_infid_set(U_dict: dict, index, dims, psi_0, proj=True):
+def state_transfer_infid_set(
+    propagators: dict, instructions: dict, index, dims, psi_0, proj=True
+):
     """
     Mean state transfer infidelity.
 
     Parameters
     ----------
-    U_dict : dict
+    propagators : dict
         Contains unitary representations of the gates, identified by a key.
     index : int
         Index of the qubit(s) in the Hilbert space to be evaluated
@@ -61,33 +68,34 @@ def state_transfer_infid_set(U_dict: dict, index, dims, psi_0, proj=True):
     Returns
     -------
     tf.float
-        State infidelity, averaged over the gates in U_dict
+        State infidelity, averaged over the gates in propagators
     """
     infids = []
-    for gate in U_dict.keys():
-        infid = state_transfer_infid(U_dict, gate, index, dims, psi_0, proj)
+    for gate, propagator in propagators.items():
+        perfect_gate = instructions[gate].get_ideal_gate(dims)
+        infid = state_transfer_infid(perfect_gate, propagator, index, dims, psi_0)
         infids.append(infid)
     return tf.reduce_mean(infids)
 
 
-def state_transfer_infid(U_dict: dict, gate: str, index, dims, psi_0, proj: bool):
+@fid_reg_deco
+def state_transfer_infid(ideal: np.array, actual: tf.constant, index, dims, psi_0):
     """
-    Single gate state transfer infidelity.
+    Single gate state transfer infidelity. The dimensions of psi_0 and ideal need to be
+    compatible and index and dims need to project actual to these same dimensions.
 
     Parameters
     ----------
-    U_dict : dict
-        Contains unitary representations of the gates, identified by a key.
+    ideal: np.array
+        Contains ideal unitary representations of the gate
+    actual: tf.Tensor
+        Contains actual unitary representations of the gate
     index : int
         Index of the qubit(s) in the Hilbert space to be evaluated
-    gate : str
-        One of the keys of U_dict, selects the gate to be evaluated
     dims : list
         List of dimensions of qubits
-    psi_0 : tf.Tensor
-        Initial state of the device
-    proj : boolean
-        Project to computational subspace
+    psi_0: tf.Tensor
+        Initial state
 
     Returns
     -------
@@ -95,72 +103,66 @@ def state_transfer_infid(U_dict: dict, gate: str, index, dims, psi_0, proj: bool
         State infidelity for the selected gate
 
     """
-    U = U_dict[gate]
-    projection = "fulluni"
-    if proj:
-        projection = "wzeros"
-    U_ideal = tf.Variable(
-        perfect_gate(gate, index, dims, projection), dtype=tf.complex128
-    )
-    psi_ideal = tf.matmul(U_ideal, psi_0)
-    psi_actual = tf.matmul(U, psi_0)
+    actual_comp = tf_project_to_comp(actual, dims=dims, index=index)
+    psi_ideal = tf.matmul(ideal, psi_0)
+    psi_actual = tf.matmul(actual_comp, psi_0)
     overlap = tf_ketket_fid(psi_ideal, psi_actual)
     infid = 1 - overlap
     return infid
 
 
 @fid_reg_deco
-def unitary_infid(U_dict: dict, gate: str, index, dims, proj: bool):
+def unitary_infid(
+    ideal: np.array, actual: tf.Tensor, index: List[int] = None, dims=None
+) -> tf.Tensor:
     """
     Unitary overlap between ideal and actually performed gate.
 
     Parameters
     ----------
-    U_dict : dict
-        Contains unitary representations of the gates, identified by a key.
-    index : int
+    ideal : np.array
+        Ideal or goal unitary representation of the gate.
+    actual : np.array
+        Actual, physical unitary representation of the gate.
+    index : List[int]
         Index of the qubit(s) in the Hilbert space to be evaluated
     gate : str
-        One of the keys of U_dict, selects the gate to be evaluated
+        One of the keys of propagators, selects the gate to be evaluated
     dims : list
         List of dimensions of qubits
-    proj : boolean
-        Project to computational subspace
 
     Returns
     -------
     tf.float
         Unitary fidelity.
     """
-    U = U_dict[gate]
-    projection = "fulluni"
-    fid_lvls = np.prod([dims[i] for i in index])
-    if proj:
-        projection = "wzeros"
-        fid_lvls = 2 ** len(index)
-    U_ideal = tf.Variable(
-        perfect_gate(gate, index, dims, projection), dtype=tf.complex128
-    )
-    infid = 1 - tf_unitary_overlap(U, U_ideal, lvls=fid_lvls)
+    if index is None:
+        index = list(range(len(dims)))
+    actual_comp = tf_project_to_comp(actual, dims=dims, index=index)
+    fid_lvls = 2 ** len(index)
+    infid = 1 - tf_unitary_overlap(actual_comp, ideal, lvls=fid_lvls)
     return infid
 
 
 @fid_reg_deco
-def unitary_infid_set(U_dict: dict, index, dims, eval, proj=True):
+def unitary_infid_set(propagators: dict, instructions: dict, index, dims, n_eval=-1):
     """
     Mean unitary overlap between ideal and actually performed gate for the gates in
-    U_dict.
+    propagators.
 
     Parameters
     ----------
-    U_dict : dict
-        Contains unitary representations of the gates, identified by a key.
-    index : int
+    propagators : dict
+        Contains actual unitary representations of the gates, resulting from physical
+        simulation
+    instructions : dict
+        Contains the perfect unitary representations of the gates, identified by a key.
+    index : List[int]
         Index of the qubit(s) in the Hilbert space to be evaluated
     dims : list
         List of dimensions of qubits
-    proj : boolean
-        Project to computational subspace
+    n_eval : int
+        Number of evaluation
 
     Returns
     -------
@@ -168,111 +170,112 @@ def unitary_infid_set(U_dict: dict, index, dims, eval, proj=True):
         Unitary fidelity.
     """
     infids = []
-    for gate in U_dict.keys():
-        infid = unitary_infid(U_dict, gate, index, dims, proj)
+    for gate, propagator in propagators.items():
+        perfect_gate = instructions[gate].get_ideal_gate(dims, index)
+        infid = unitary_infid(perfect_gate, propagator, index, dims)
         infids.append(infid)
     return tf.reduce_mean(infids)
 
 
 @fid_reg_deco
-def lindbladian_unitary_infid(U_dict: dict, gate: str, index, dims, proj: bool):
+def lindbladian_unitary_infid(
+    ideal: np.array, actual: tf.constant, index=[0], dims=[2]
+) -> tf.constant:
     """
     Variant of the unitary fidelity for the Lindbladian propagator.
 
     Parameters
     ----------
-    U_dict : dict
-        Contains unitary representations of the gates, identified by a key.
-    index : int
+    ideal: np.array
+        Contains ideal unitary representations of the gate
+    actual: tf.Tensor
+        Contains actual unitary representations of the gate
+    index : List[int]
         Index of the qubit(s) in the Hilbert space to be evaluated
-    gate : str
-        One of the keys of U_dict, selects the gate to be evaluated
     dims : list
         List of dimensions of qubits
-    proj : boolean
-        Project to computational subspace
 
     Returns
     -------
     tf.float
         Overlap fidelity for the Lindblad propagator.
     """
-    # Here we deal with the projected case differently because it's not easy
-    # to select the right section of the superoper
-    U = U_dict[gate]
-    projection = "fulluni"
-    fid_lvls = np.prod([dims[i] for i in index])
-    if proj:
-        projection = "wzeros"
-        fid_lvls = 2 ** len(index)
-    U_ideal = tf_super(
-        tf.Variable(perfect_gate(gate, index, dims, projection), dtype=tf.complex128)
-    )
-    infid = 1 - tf_superoper_unitary_overlap(U, U_ideal, lvls=fid_lvls)
+    U_ideal = tf_super(ideal)
+    actual_comp = tf_project_to_comp(actual, dims=dims, index=index, to_super=True)
+    fid_lvls = 2 ** len(index)
+    infid = 1 - tf_superoper_unitary_overlap(actual_comp, U_ideal, lvls=fid_lvls)
     return infid
 
 
 @fid_reg_deco
-def lindbladian_unitary_infid_set(U_dict: dict, index, dims, eval, proj=True):
+def lindbladian_unitary_infid_set(
+    propagators: dict, instructions: Dict[str, Instruction], index, dims, n_eval
+):
     """
     Variant of the mean unitary fidelity for the Lindbladian propagator.
 
     Parameters
     ----------
-    U_dict : dict
-        Contains unitary representations of the gates, identified by a key.
-    index : int
+    propagators : dict
+        Contains actual unitary representations of the gates, resulting from physical
+        simulation
+    instructions : dict
+        Contains the perfect unitary representations of the gates, identified by a key.
+    index : List[int]
         Index of the qubit(s) in the Hilbert space to be evaluated
     dims : list
         List of dimensions of qubits
-    proj : boolean
-        Project to computational subspace
+    n_eval : int
+        Number of evaluation
 
     Returns
     -------
     tf.float
-        Mean overlap fidelity for the Lindblad propagator for all gates in U_dict.
+        Mean overlap fidelity for the Lindblad propagator for all gates in propagators.
     """
     infids = []
-    for gate in U_dict.keys():
-        infid = lindbladian_unitary_infid(U_dict, gate, index, dims, proj)
+    for gate, propagator in propagators.items():
+        perfect_gate = instructions[gate].get_ideal_gate(dims)
+        infid = lindbladian_unitary_infid(perfect_gate, propagator, index, dims)
         infids.append(infid)
     return tf.reduce_mean(infids)
 
 
 @fid_reg_deco
-def average_infid(U_dict: dict, gate: str, index, dims, proj=True):
+def average_infid(
+    ideal: np.array, actual: tf.Tensor, index: List[int] = [0], dims=[2]
+) -> tf.constant:
     """
     Average fidelity uses the Pauli basis to compare. Thus, perfect gates are
     always 2x2 (per qubit) and the actual unitary needs to be projected down.
 
     Parameters
     ----------
-    U_dict : dict
-        Contains unitary representations of the gates, identified by a key.
-    index : int
+    ideal: np.array
+        Contains ideal unitary representations of the gate
+    actual: tf.Tensor
+        Contains actual unitary representations of the gate
+    index : List[int]
         Index of the qubit(s) in the Hilbert space to be evaluated
     dims : list
         List of dimensions of qubits
-    proj : boolean
-        Project to computational subspace
     """
-    U = U_dict[gate]
-    U_ideal = tf.Variable(
-        perfect_gate(gate, index, dims=[2] * len(dims)), dtype=tf.complex128
-    )
-    infid = 1 - tf_average_fidelity(U, U_ideal, lvls=dims)
+    actual_comp = tf_project_to_comp(actual, dims=dims, index=index)
+    fid_lvls = [2] * len(index)
+    infid = 1 - tf_average_fidelity(actual_comp, ideal, lvls=fid_lvls)
     return infid
 
 
 @fid_reg_deco
-def average_infid_set(U_dict: dict, index, dims, eval, proj=True):
+def average_infid_set(
+    propagators: dict, instructions: dict, index: List[int], dims, n_eval=-1
+):
     """
-    Mean average fidelity over all gates in U_dict.
+    Mean average fidelity over all gates in propagators.
 
     Parameters
     ----------
-    U_dict : dict
+    propagators : dict
         Contains unitary representations of the gates, identified by a key.
     index : int
         Index of the qubit(s) in the Hilbert space to be evaluated
@@ -287,20 +290,21 @@ def average_infid_set(U_dict: dict, index, dims, eval, proj=True):
         Mean average fidelity
     """
     infids = []
-    for gate in U_dict.keys():
-        infid = average_infid(U_dict, gate, index, dims, proj)
+    for gate, propagator in propagators.items():
+        perfect_gate = instructions[gate].get_ideal_gate(dims, index)
+        infid = average_infid(perfect_gate, propagator, index, dims)
         infids.append(infid)
     return tf.reduce_mean(infids)
 
 
 @fid_reg_deco
-def average_infid_seq(U_dict: dict, index, dims, eval, proj=True):
+def average_infid_seq(propagators: dict, instructions: dict, index, dims, n_eval=-1):
     """
-    Average sequence fidelity over all gates in U_dict.
+    Average sequence fidelity over all gates in propagators.
 
     Parameters
     ----------
-    U_dict : dict
+    propagators : dict
         Contains unitary representations of the gates, identified by a key.
     index : int
         Index of the qubit(s) in the Hilbert space to be evaluated
@@ -315,45 +319,47 @@ def average_infid_seq(U_dict: dict, index, dims, eval, proj=True):
         Mean average fidelity
     """
     fid = 1
-    for gate in U_dict.keys():
-        fid *= 1 - average_infid(U_dict, gate, index, dims, proj)
+    for gate, propagator in propagators.items():
+        perfect_gate = instructions[gate].get_ideal_gate(dims)
+        fid *= 1 - average_infid(perfect_gate, propagator, index, dims)
     return 1 - fid
 
 
 @fid_reg_deco
-def lindbladian_average_infid(U_dict: dict, gate: str, index, dims, proj=True):
+def lindbladian_average_infid(
+    ideal: np.array, actual: tf.constant, index=[0], dims=[2]
+) -> tf.constant:
     """
     Average fidelity uses the Pauli basis to compare. Thus, perfect gates are
     always 2x2 (per qubit) and the actual unitary needs to be projected down.
 
     Parameters
     ----------
-    U_dict : dict
-        Contains unitary representations of the gates, identified by a key.
+    ideal: np.array
+        Contains ideal unitary representations of the gate
+    actual: tf.Tensor
+        Contains actual unitary representations of the gate
     index : int
         Index of the qubit(s) in the Hilbert space to be evaluated
     dims : list
         List of dimensions of qubits
-    proj : boolean
-        Project to computational subspace
     """
-    U = U_dict[gate]
-    ideal = tf.Variable(
-        perfect_gate(gate, index, dims=[2] * len(dims)), dtype=tf.complex128
-    )
     U_ideal = tf_super(ideal)
-    infid = 1 - tf_superoper_average_fidelity(U, U_ideal, lvls=dims)
+    actual_comp = tf_project_to_comp(actual, dims=dims, index=index, to_super=True)
+    infid = 1 - tf_superoper_average_fidelity(actual_comp, U_ideal, lvls=dims)
     return infid
 
 
 @fid_reg_deco
-def lindbladian_average_infid_set(U_dict: dict, index, dims, eval, proj=True):
+def lindbladian_average_infid_set(
+    propagators: dict, instructions: Dict[str, Instruction], index, dims, n_eval
+):
     """
-    Mean average fidelity over all gates in U_dict.
+    Mean average fidelity over all gates in propagators.
 
     Parameters
     ----------
-    U_dict : dict
+    propagators : dict
         Contains unitary representations of the gates, identified by a key.
     index : int
         Index of the qubit(s) in the Hilbert space to be evaluated
@@ -368,22 +374,25 @@ def lindbladian_average_infid_set(U_dict: dict, index, dims, eval, proj=True):
         Mean average fidelity
     """
     infids = []
-    for gate in U_dict.keys():
-        infid = lindbladian_average_infid(U_dict, gate, index, dims, proj)
+    for gate, propagator in propagators.items():
+        perfect_gate = instructions[gate].get_ideal_gate(dims)
+        infid = lindbladian_average_infid(perfect_gate, propagator, index, dims)
         infids.append(infid)
     return tf.reduce_mean(infids)
 
 
 @fid_reg_deco
-def epc_analytical(U_dict: dict, index, dims, proj: bool, cliffords=False):
+def epc_analytical(propagators: dict, index, dims, proj: bool, cliffords=False):
     # TODO check this work with new index and dims (double-check)
     num_gates = len(dims)
     if cliffords:
-        real_cliffords = evaluate_sequences(U_dict, [[C] for C in cliffords_string])
+        real_cliffords = evaluate_sequences(
+            propagators, [[C] for C in cliffords_string]
+        )
     elif num_gates == 1:
-        real_cliffords = evaluate_sequences(U_dict, cliffords_decomp)
+        real_cliffords = evaluate_sequences(propagators, cliffords_decomp)
     elif num_gates == 2:
-        real_cliffords = evaluate_sequences(U_dict, cliffords_decomp_xId)
+        real_cliffords = evaluate_sequences(propagators, cliffords_decomp_xId)
     ideal_cliffords = perfect_cliffords(lvls=[2] * num_gates, num_gates=num_gates)
     fids = []
     for C_indx in range(24):
@@ -396,14 +405,18 @@ def epc_analytical(U_dict: dict, index, dims, proj: bool, cliffords=False):
 
 
 @fid_reg_deco
-def lindbladian_epc_analytical(U_dict: dict, index, dims, proj: bool, cliffords=False):
+def lindbladian_epc_analytical(
+    propagators: dict, index, dims, proj: bool, cliffords=False
+):
     num_gates = len(dims)
     if cliffords:
-        real_cliffords = evaluate_sequences(U_dict, [[C] for C in cliffords_string])
+        real_cliffords = evaluate_sequences(
+            propagators, [[C] for C in cliffords_string]
+        )
     elif num_gates == 1:
-        real_cliffords = evaluate_sequences(U_dict, cliffords_decomp)
+        real_cliffords = evaluate_sequences(propagators, cliffords_decomp)
     elif num_gates == 2:
-        real_cliffords = evaluate_sequences(U_dict, cliffords_decomp_xId)
+        real_cliffords = evaluate_sequences(propagators, cliffords_decomp_xId)
     ideal_cliffords = perfect_cliffords(lvls=[2] * num_gates, num_gates=num_gates)
     fids = []
     for C_indx in range(24):
@@ -429,18 +442,18 @@ def populations(state, lindbladian):
 
 
 @fid_reg_deco
-def population(U_dict: dict, lvl: int, gate: str):
-    U = U_dict[gate]
+def population(propagators: dict, lvl: int, gate: str):
+    U = propagators[gate]
     lvls = U.shape[0]
-    psi_0 = tf.Variable(basis(lvls, 0), dtype=tf.complex128)
+    psi_0 = tf.constant(basis(lvls, 0), dtype=tf.complex128)
     psi_actual = tf.matmul(U, psi_0)
     return populations(psi_actual, lindbladian=False)[lvl]
 
 
-def lindbladian_population(U_dict: dict, lvl: int, gate: str):
-    U = U_dict[gate]
+def lindbladian_population(propagators: dict, lvl: int, gate: str):
+    U = propagators[gate]
     lvls = int(np.sqrt(U.shape[0]))
-    psi_0 = tf.Variable(basis(lvls, 0), dtype=tf.complex128)
+    psi_0 = tf.constant(basis(lvls, 0), dtype=tf.complex128)
     dv_0 = tf_dm_to_vec(tf_state_to_dm(psi_0))
     dv_actual = tf.matmul(U, dv_0)
     return populations(dv_actual, lindbladian=True)[lvl]
@@ -448,7 +461,7 @@ def lindbladian_population(U_dict: dict, lvl: int, gate: str):
 
 @fid_reg_deco
 def RB(
-    U_dict,
+    propagators,
     min_length: int = 5,
     max_length: int = 500,
     num_lengths: int = 20,
@@ -457,10 +470,10 @@ def RB(
     lindbladian=False,
     padding="",
 ):
-    gate = list(U_dict.keys())[0]
-    U = U_dict[gate]
+    gate = list(propagators.keys())[0]
+    U = propagators[gate]
     dim = int(U.shape[0])
-    psi_init = tf.Variable(basis(dim, 0), dtype=tf.complex128)
+    psi_init = tf.constant(basis(dim, 0), dtype=tf.complex128)
     if logspace:
         lengths = np.rint(
             np.logspace(np.log10(min_length), np.log10(max_length), num=num_lengths)
@@ -472,7 +485,7 @@ def RB(
     surv_prob = []
     for L in lengths:
         seqs = single_length_RB(num_seqs, L, padding)
-        Us = evaluate_sequences(U_dict, seqs)
+        Us = evaluate_sequences(propagators, seqs)
         pop0s = []
         for U in Us:
             pops = populations(tf.matmul(U, psi_init), lindbladian)
@@ -513,7 +526,7 @@ def RB(
             max_length = max_length * 2
             for L in new_lengths:
                 seqs = single_length_RB(num_seqs, L, padding)
-                Us = evaluate_sequences(U_dict, seqs)
+                Us = evaluate_sequences(propagators, seqs)
                 pop0s = []
                 for U in Us:
                     pops = populations(tf.matmul(U, psi_init), lindbladian)
@@ -527,23 +540,23 @@ def RB(
 
 @fid_reg_deco
 def lindbladian_RB_left(
-    U_dict: dict,
+    propagators: dict,
     gate: str,
     index,
     dims,
     proj: bool = False,
 ):
-    return RB(U_dict, padding="left")
+    return RB(propagators, padding="left")
 
 
 @fid_reg_deco
-def lindbladian_RB_right(U_dict: dict, gate: str, index, dims, proj: bool):
-    return RB(U_dict, padding="right")
+def lindbladian_RB_right(propagators: dict, gate: str, index, dims, proj: bool):
+    return RB(propagators, padding="right")
 
 
 @fid_reg_deco
 def leakage_RB(
-    U_dict,
+    propagators,
     min_length: int = 5,
     max_length: int = 500,
     num_lengths: int = 20,
@@ -551,10 +564,10 @@ def leakage_RB(
     logspace=False,
     lindbladian=False,
 ):
-    gate = list(U_dict.keys())[0]
-    U = U_dict[gate]
+    gate = list(propagators.keys())[0]
+    U = propagators[gate]
     dim = int(U.shape[0])
-    psi_init = tf.Variable(basis(dim, 0), dtype=tf.complex128)
+    psi_init = tf.constant(basis(dim, 0), dtype=tf.complex128)
     if logspace:
         lengths = np.rint(
             np.logspace(np.log10(min_length), np.log10(max_length), num=num_lengths)
@@ -567,7 +580,7 @@ def leakage_RB(
     surv_prob = []
     for L in lengths:
         seqs = single_length_RB(num_seqs, L)
-        Us = evaluate_sequences(U_dict, seqs)
+        Us = evaluate_sequences(propagators, seqs)
         pop0s = []
         pop_comps = []
         for U in Us:
@@ -616,7 +629,7 @@ def leakage_RB(
             max_length = max_length * 2
             for L in new_lengths:
                 seqs = single_length_RB(num_seqs, L)
-                Us = evaluate_sequences(U_dict, seqs)
+                Us = evaluate_sequences(propagators, seqs)
                 pop0s = []
                 pop_comps = []
                 for U in Us:
@@ -667,7 +680,7 @@ def leakage_RB(
             max_length = max_length * 2
             for L in new_lengths:
                 seqs = single_length_RB(num_seqs, L)
-                Us = evaluate_sequences(U_dict, seqs)
+                Us = evaluate_sequences(propagators, seqs)
                 pop0s = []
                 pop_comps = []
                 for U in Us:
@@ -687,7 +700,7 @@ def leakage_RB(
 
 @fid_reg_deco
 def orbit_infid(
-    U_dict,
+    propagators,
     RB_number: int = 30,
     RB_length: int = 20,
     lindbladian=False,
@@ -697,11 +710,11 @@ def orbit_infid(
 ):
     if not seqs:
         seqs = single_length_RB(RB_number=RB_number, RB_length=RB_length)
-    Us = evaluate_sequences(U_dict, seqs)
+    Us = evaluate_sequences(propagators, seqs)
     infids = []
     for U in Us:
         dim = int(U.shape[0])
-        psi_init = tf.Variable(basis(dim, 0), dtype=tf.complex128)
+        psi_init = tf.constant(basis(dim, 0), dtype=tf.complex128)
         psi_actual = tf.matmul(U, psi_init)
         pop0 = tf_abs(psi_actual[0]) ** 2
         p1 = 1 - pop0
