@@ -3,7 +3,15 @@
 import pickle
 import numpy as np
 import pytest
-from c3.generator.devices import LO, AWG, Mixer, Response, DigitalToAnalog, VoltsToHertz
+from c3.generator.devices import (
+    LO,
+    AWG,
+    Mixer,
+    Response,
+    DigitalToAnalog,
+    VoltsToHertz,
+    Crosstalk,
+)
 from c3.generator.generator import Generator
 from c3.signal.gates import Instruction
 from c3.signal.pulse import Envelope, Carrier
@@ -30,6 +38,14 @@ v_to_hz = VoltsToHertz(
     inputs=1,
     outputs=1,
 )
+xtalk = Crosstalk(
+    name="crosstalk",
+    channels=["d1", "d2"],
+    crosstalk_matrix=Quantity(
+        value=[[1, 0], [1, 0]], min_val=[[0, 0], [0, 0]], max_val=[[1, 1], [1, 1]]
+    ),
+)
+
 generator = Generator(
     devices={
         "LO": lo,
@@ -79,12 +95,12 @@ carr = Carrier(
     name="carrier", desc="Frequency of the local oscillator", params=carrier_parameters
 )
 
-X90p_q1 = Instruction(name="X90p", t_start=0.0, t_end=t_final, channels=["d1"])
-X90p_q1.add_component(gauss_env_single, "d1")
-X90p_q1.add_component(carr, "d1")
+rx90p_q1 = Instruction(name="rx90p", t_start=0.0, t_end=t_final, channels=["d1"])
+rx90p_q1.add_component(gauss_env_single, "d1")
+rx90p_q1.add_component(carr, "d1")
 
-tstart = X90p_q1.t_start
-tend = X90p_q1.t_end
+tstart = rx90p_q1.t_start
+tend = rx90p_q1.t_end
 chan = "d1"
 
 with open("test/generator_data.pickle", "rb") as filename:
@@ -93,7 +109,7 @@ with open("test/generator_data.pickle", "rb") as filename:
 
 @pytest.mark.unit
 def test_LO() -> None:
-    lo_sig = lo.process(X90p_q1, "d1")
+    lo_sig = lo.process(rx90p_q1, "d1")
     assert (
         lo_sig["inphase"].numpy() - data["lo_sig"]["values"][0].numpy() < 1e-12
     ).all()
@@ -105,7 +121,7 @@ def test_LO() -> None:
 
 @pytest.mark.unit
 def test_AWG() -> None:
-    awg_sig = awg.process(X90p_q1, "d1")
+    awg_sig = awg.process(rx90p_q1, "d1")
     assert (
         awg_sig["inphase"].numpy() - data["awg_sig"]["inphase"].numpy() < 1e-12
     ).all()
@@ -116,7 +132,7 @@ def test_AWG() -> None:
 
 @pytest.mark.unit
 def test_DAC() -> None:
-    dac_sig = dac.process(X90p_q1, "d1", data["awg_sig"])
+    dac_sig = dac.process(rx90p_q1, "d1", data["awg_sig"])
     assert (
         dac_sig["inphase"].numpy() - data["dig_to_an_sig"]["inphase"].numpy() < 1e-12
     ).all()
@@ -128,7 +144,7 @@ def test_DAC() -> None:
 
 @pytest.mark.unit
 def test_Response() -> None:
-    resp_sig = resp.process(X90p_q1, "d1", data["dig_to_an_sig"])
+    resp_sig = resp.process(rx90p_q1, "d1", data["dig_to_an_sig"])
     assert (
         resp_sig["inphase"].numpy() - data["resp_sig"]["inphase"].numpy() < 1e-12
     ).all()
@@ -145,22 +161,75 @@ def test_mixer() -> None:
         "quadrature": data["lo_sig"]["values"][1],
         "ts": data["lo_sig"]["ts"],
     }
-    mixed_sig = mixer.process(X90p_q1, "d1", lo_signal, data["resp_sig"])
+    mixed_sig = mixer.process(rx90p_q1, "d1", lo_signal, data["resp_sig"])
     assert (mixed_sig["values"].numpy() - data["mixer_sig"].numpy() < 1e-12).all()
 
 
 @pytest.mark.unit
 def test_v2hz() -> None:
     mixer_sig = {"values": data["mixer_sig"], "ts": data["lo_sig"]["ts"]}
-    final_sig = v_to_hz.process(X90p_q1, "d1", mixer_sig)
+    final_sig = v_to_hz.process(rx90p_q1, "d1", mixer_sig)
     assert (final_sig["values"].numpy() - data["v2hz_sig"].numpy() < 1).all()
 
 
 @pytest.mark.integration
 def test_full_signal_chain() -> None:
-    full_signal = generator.generate_signals(X90p_q1)
+    full_signal = generator.generate_signals(rx90p_q1)
     assert (
         full_signal["d1"]["values"].numpy()
         - data["full_signal"][0]["d1"]["values"].numpy()
         < 1
     ).all()
+
+
+@pytest.mark.integration
+def test_crosstalk() -> None:
+    generator = Generator(
+        devices={
+            "LO": lo,
+            "AWG": awg,
+            "DigitalToAnalog": dac,
+            "Response": resp,
+            "Mixer": mixer,
+            "VoltsToHertz": v_to_hz,
+            "crosstalk": xtalk,
+        },
+        chains={
+            "d1": ["LO", "AWG", "DigitalToAnalog", "Response", "Mixer", "VoltsToHertz"],
+            "d2": ["LO", "AWG", "DigitalToAnalog", "Response", "Mixer", "VoltsToHertz"],
+        },
+    )
+    RX90p_q1 = Instruction(
+        name="RX90p", t_start=0.0, t_end=t_final, channels=["d1", "d2"]
+    )
+    RX90p_q1.add_component(gauss_env_single, "d1")
+    RX90p_q1.add_component(carr, "d1")
+
+    gauss_params_single_2 = {
+        "amp": Quantity(value=0, min_val=-0.4, max_val=0.6, unit="V"),
+        "t_final": Quantity(
+            value=t_final, min_val=0.5 * t_final, max_val=1.5 * t_final, unit="s"
+        ),
+        "sigma": Quantity(
+            value=t_final / 4, min_val=t_final / 8, max_val=t_final / 2, unit="s"
+        ),
+        "xy_angle": Quantity(
+            value=0.0, min_val=-0.5 * np.pi, max_val=2.5 * np.pi, unit="rad"
+        ),
+        "freq_offset": Quantity(
+            value=-sideband - 3e6, min_val=-56 * 1e6, max_val=-52 * 1e6, unit="Hz 2pi"
+        ),
+        "delta": Quantity(value=-1, min_val=-5, max_val=3, unit=""),
+    }
+    gauss_env_single_2 = Envelope(
+        name="gauss",
+        desc="Gaussian comp for single-qubit gates",
+        params=gauss_params_single_2,
+        shape=env_lib.gaussian_nonorm,
+    )
+    RX90p_q1.add_component(gauss_env_single_2, "d2")
+    RX90p_q1.add_component(carr, "d2")
+    full_signal = generator.generate_signals(RX90p_q1)
+    assert (
+        full_signal["d1"]["values"].numpy() == full_signal["d2"]["values"].numpy()
+    ).all()
