@@ -6,7 +6,7 @@ import pytest
 from c3.experiment import Experiment
 
 exp = Experiment()
-exp.quick_setup("test/quickstart.hjson")
+exp.load_quick_setup("test/quickstart.hjson")
 pmap = exp.pmap
 model = pmap.model
 generator = pmap.generator
@@ -17,8 +17,9 @@ def test_exp_quick_setup_freqs() -> None:
     """
     Test the quick setup.
     """
+    print(pmap.instructions.keys())
     qubit_freq = model.subsystems["Q1"].params["freq"].get_value()
-    gate = pmap.instructions["X90p:Id"]
+    gate = pmap.instructions["rx90p[0]"]
     carrier_freq = gate.comps["d1"]["carrier"].params["freq"].get_value()
     offset = gate.comps["d1"]["gaussian"].params["freq_offset"].get_value()
     assert qubit_freq == carrier_freq + offset
