@@ -2,34 +2,37 @@
 integration testing module for C1 optimization through two-qubits example
 """
 
+import os
+import tempfile
 import copy
 import pickle
 import numpy as np
-from typing import List
 import pytest
+from numpy.testing import assert_array_almost_equal as almost_equal
 
 # Main C3 objects
 from c3.c3objs import Quantity as Qty
 from c3.parametermap import ParameterMap as Pmap
 from c3.experiment import Experiment as Exp
-from c3.system.model import Model as Mdl
+from c3.model import Model as Mdl
 from c3.generator.generator import Generator as Gnr
 
 # Building blocks
 import c3.generator.devices as devices
-import c3.system.chip as chip
 import c3.signal.pulse as pulse
 import c3.signal.gates as gates
-import c3.system.tasks as tasks
 
 # Libs and helpers
 import c3.libraries.algorithms as algorithms
-import c3.libraries.hamiltonians as hamiltonians
-import c3.libraries.fidelities as fidelities
+import c3.libraries.chip as chip
 import c3.libraries.envelopes as envelopes
+import c3.libraries.fidelities as fidelities
+import c3.libraries.hamiltonians as hamiltonians
+import c3.libraries.tasks as tasks
 
-from c3.optimizers.c1 import C1
+from c3.optimizers.optimalcontrol import OptimalControl
 
+logdir = os.path.join(tempfile.TemporaryDirectory().name, "c3logs")
 
 qubit_lvls = 3
 freq_q1 = 5e9
@@ -175,12 +178,6 @@ gauss_env_single = pulse.Envelope(
     shape=envelopes.gaussian_nonorm,
 )
 
-lo_freq_q1 = 5e9 + sideband
-carrier_parameters = {
-    "freq": Qty(value=lo_freq_q1, min_val=4.5e9, max_val=6e9, unit="Hz 2pi"),
-    "framechange": Qty(value=0.0, min_val=-np.pi, max_val=3 * np.pi, unit="rad"),
-}
-
 nodrive_env = pulse.Envelope(
     name="no_drive",
     params={
@@ -196,6 +193,7 @@ carrier_parameters = {
     "freq": Qty(value=lo_freq_q1, min_val=4.5e9, max_val=6e9, unit="Hz 2pi"),
     "framechange": Qty(value=0.0, min_val=-np.pi, max_val=3 * np.pi, unit="rad"),
 }
+
 carr = pulse.Carrier(
     name="carrier", desc="Frequency of the local oscillator", params=carrier_parameters
 )
@@ -204,99 +202,120 @@ lo_freq_q2 = 5.6e9 + sideband
 carr_2 = copy.deepcopy(carr)
 carr_2.params["freq"].set_value(lo_freq_q2)
 
-X90p_q1 = gates.Instruction(name="X90p", t_start=0.0, t_end=t_final, channels=["d1"])
-X90p_q2 = gates.Instruction(name="X90p", t_start=0.0, t_end=t_final, channels=["d2"])
-QId_q1 = gates.Instruction(name="Id", t_start=0.0, t_end=t_final, channels=["d1"])
-QId_q2 = gates.Instruction(name="Id", t_start=0.0, t_end=t_final, channels=["d2"])
+rx90p_q1 = gates.Instruction(
+    name="rx90p",
+    targets=[0],
+    t_start=0.0,
+    t_end=t_final,
+    channels=["d1", "d2"],
+    params={"use_t_before": True},
+)
+rx90p_q2 = gates.Instruction(
+    name="rx90p",
+    targets=[1],
+    t_start=0.0,
+    t_end=t_final,
+    channels=["d1", "d2"],
+    params={"use_t_before": True},
+)
+QId_q1 = gates.Instruction(
+    name="id",
+    targets=[0],
+    t_start=0.0,
+    t_end=t_final,
+    channels=["d1", "d2"],
+    params={"use_t_before": True},
+)
+QId_q2 = gates.Instruction(
+    name="id",
+    targets=[1],
+    t_start=0.0,
+    t_end=t_final,
+    channels=["d1", "d2"],
+    params={"use_t_before": True},
+)
+
+rx90p_q1.add_component(gauss_env_single, "d1")
+rx90p_q1.add_component(carr, "d1")
+rx90p_q1.add_component(nodrive_env, "d2")
+rx90p_q1.add_component(copy.deepcopy(carr_2), "d2")
+rx90p_q1.comps["d2"]["carrier"].params["framechange"].set_value(
+    (-sideband * t_final) * 2 * np.pi % (2 * np.pi)
+)
+
+rx90p_q2.add_component(copy.deepcopy(gauss_env_single), "d2")
+rx90p_q2.add_component(carr_2, "d2")
+rx90p_q2.add_component(nodrive_env, "d1")
+rx90p_q2.add_component(copy.deepcopy(carr), "d1")
+rx90p_q2.comps["d1"]["carrier"].params["framechange"].set_value(
+    (-sideband * t_final) * 2 * np.pi % (2 * np.pi)
+)
+
 
-X90p_q1.add_component(gauss_env_single, "d1")
-X90p_q1.add_component(carr, "d1")
 QId_q1.add_component(nodrive_env, "d1")
 QId_q1.add_component(copy.deepcopy(carr), "d1")
-
-X90p_q2.add_component(copy.deepcopy(gauss_env_single), "d2")
-X90p_q2.add_component(carr_2, "d2")
+QId_q1.add_component(nodrive_env, "d2")
+QId_q1.add_component(copy.deepcopy(carr_2), "d2")
 QId_q2.add_component(copy.deepcopy(nodrive_env), "d2")
 QId_q2.add_component(copy.deepcopy(carr_2), "d2")
+QId_q2.add_component(nodrive_env, "d1")
+QId_q2.add_component(copy.deepcopy(carr), "d1")
 
-QId_q1.comps["d1"]["carrier"].params["framechange"].set_value(
-    (-sideband * t_final) * 2 * np.pi % (2 * np.pi)
-)
-QId_q2.comps["d2"]["carrier"].params["framechange"].set_value(
-    (-sideband * t_final) * 2 * np.pi % (2 * np.pi)
-)
-
-Y90p_q1 = copy.deepcopy(X90p_q1)
-Y90p_q1.name = "Y90p"
-X90m_q1 = copy.deepcopy(X90p_q1)
-X90m_q1.name = "X90m"
-Y90m_q1 = copy.deepcopy(X90p_q1)
-Y90m_q1.name = "Y90m"
+Y90p_q1 = copy.deepcopy(rx90p_q1)
+Y90p_q1.name = "ry90p"
+X90m_q1 = copy.deepcopy(rx90p_q1)
+X90m_q1.name = "rx90m"
+Y90m_q1 = copy.deepcopy(rx90p_q1)
+Y90m_q1.name = "ry90m"
 Y90p_q1.comps["d1"]["gauss"].params["xy_angle"].set_value(0.5 * np.pi)
 X90m_q1.comps["d1"]["gauss"].params["xy_angle"].set_value(np.pi)
 Y90m_q1.comps["d1"]["gauss"].params["xy_angle"].set_value(1.5 * np.pi)
-Q1_gates = [QId_q1, X90p_q1, Y90p_q1, X90m_q1, Y90m_q1]
+single_q_gates = [QId_q1, rx90p_q1, Y90p_q1, X90m_q1, Y90m_q1]
 
 
-Y90p_q2 = copy.deepcopy(X90p_q2)
-Y90p_q2.name = "Y90p"
-X90m_q2 = copy.deepcopy(X90p_q2)
-X90m_q2.name = "X90m"
-Y90m_q2 = copy.deepcopy(X90p_q2)
-Y90m_q2.name = "Y90m"
+Y90p_q2 = copy.deepcopy(rx90p_q2)
+Y90p_q2.name = "ry90p"
+X90m_q2 = copy.deepcopy(rx90p_q2)
+X90m_q2.name = "rx90m"
+Y90m_q2 = copy.deepcopy(rx90p_q2)
+Y90m_q2.name = "ry90m"
 Y90p_q2.comps["d2"]["gauss"].params["xy_angle"].set_value(0.5 * np.pi)
 X90m_q2.comps["d2"]["gauss"].params["xy_angle"].set_value(np.pi)
 Y90m_q2.comps["d2"]["gauss"].params["xy_angle"].set_value(1.5 * np.pi)
-Q2_gates = [QId_q2, X90p_q2, Y90p_q2, X90m_q2, Y90m_q2]
+single_q_gates.extend([QId_q2, rx90p_q2, Y90p_q2, X90m_q2, Y90m_q2])
 
-all_1q_gates_comb = []
-for g1 in Q1_gates:
-    for g2 in Q2_gates:
-        g = gates.Instruction(name="NONE", t_start=0.0, t_end=t_final, channels=[])
-        g.name = g1.name + ":" + g2.name
-        channels: List[str] = []
-        channels.extend(g1.comps.keys())
-        channels.extend(g2.comps.keys())
-        for chan in channels:
-            g.comps[chan] = {}
-            if chan in g1.comps:
-                g.comps[chan].update(g1.comps[chan])
-            if chan in g2.comps:
-                g.comps[chan].update(g2.comps[chan])
-        all_1q_gates_comb.append(g)
-
-pmap = Pmap(all_1q_gates_comb, generator, model)
+pmap = Pmap(single_q_gates, generator, model)
 
 exp = Exp(pmap)
 
 generator.devices["AWG"].enable_drag_2()
 
-exp.set_opt_gates(["X90p:Id"])
+exp.set_opt_gates(["rx90p[0]"])
 
 gateset_opt_map = [
     [
-        ("X90p:Id", "d1", "gauss", "amp"),
+        ("rx90p[0]", "d1", "gauss", "amp"),
     ],
     [
-        ("X90p:Id", "d1", "gauss", "freq_offset"),
+        ("rx90p[0]", "d1", "gauss", "freq_offset"),
     ],
     [
-        ("X90p:Id", "d1", "gauss", "xy_angle"),
+        ("rx90p[0]", "d1", "gauss", "xy_angle"),
     ],
     [
-        ("X90p:Id", "d1", "gauss", "delta"),
+        ("rx90p[0]", "d1", "gauss", "delta"),
     ],
 ]
 
 pmap.set_opt_map(gateset_opt_map)
 
-opt = C1(
-    dir_path="/tmp/c3log/",
+opt = OptimalControl(
+    dir_path=logdir,
     fid_func=fidelities.average_infid_set,
     fid_subspace=["Q1", "Q2"],
     pmap=pmap,
     algorithm=algorithms.tf_sgd,
-    options={"maxfun": 2},
+    options={"maxiters": 5},
     run_name="better_X90_tf_sgd",
 )
 
@@ -305,28 +324,28 @@ opt.set_exp(exp)
 with open("test/two_qubit_data.pickle", "rb") as filename:
     test_data = pickle.load(filename)
 
-gen_signal = generator.generate_signals(pmap.instructions["X90p:Id"])
+gen_signal = generator.generate_signals(pmap.instructions["rx90p[0]"])
 ts = gen_signal["d1"]["ts"]
 hdrift, hks = model.get_Hamiltonians()
-propagator = exp.propagation(gen_signal, "X90p:Id")
+propagator = exp.propagation(gen_signal, "rx90p[0]")
 
 
 def test_signals() -> None:
-    assert (
-        gen_signal["d1"]["values"].numpy() - test_data["signal"]["d1"]["values"].numpy()
-        < 1
-    ).all()
-    assert (ts.numpy() == test_data["ts"].numpy()).all()
+    np.testing.assert_allclose(ts, test_data["ts"])
+    np.testing.assert_allclose(
+        actual=gen_signal["d1"]["values"].numpy(),
+        desired=test_data["signal"]["d1"]["values"].numpy(),
+    )
 
 
 def test_hamiltonians() -> None:
     assert (hdrift.numpy() - test_data["hdrift"].numpy() < 1).any()
     for key in hks:
-        assert (hks[key].numpy() - test_data["hks"][key].numpy() < 1).all()
+        almost_equal(hks[key], test_data["hks"][key])
 
 
 def test_propagation() -> None:
-    assert (propagator.numpy() - test_data["propagator"].numpy() < 1e-12).all()
+    almost_equal(propagator, test_data["propagator"])
 
 
 @pytest.mark.slow
@@ -337,16 +356,37 @@ def test_optim_tf_sgd() -> None:
     """
     check if optimization result is below 1e-2
     """
+    opt.store_unitaries = True
     opt.optimize_controls()
     assert opt.current_best_goal < 0.01
 
 
+@pytest.mark.optimizers
+@pytest.mark.slow
+@pytest.mark.integration
+@pytest.mark.tensorflow
+def test_bad_tf_sgd() -> None:
+    bad_tf_opt = OptimalControl(
+        dir_path=logdir,
+        fid_func=fidelities.average_infid_set,
+        fid_subspace=["Q1", "Q2"],
+        pmap=pmap,
+        algorithm=algorithms.tf_sgd,
+        options={"maxfun": 2},
+        run_name="better_X90_bad_tf",
+    )
+    bad_tf_opt.set_exp(exp)
+
+    with pytest.raises(KeyError):
+        bad_tf_opt.optimize_controls()
+
+
 @pytest.mark.optimizers
 @pytest.mark.slow
 @pytest.mark.integration
 def test_optim_lbfgs() -> None:
-    lbfgs_opt = C1(
-        dir_path="/tmp/c3log/",
+    lbfgs_opt = OptimalControl(
+        dir_path=logdir,
         fid_func=fidelities.average_infid_set,
         fid_subspace=["Q1", "Q2"],
         pmap=pmap,
@@ -358,3 +398,22 @@ def test_optim_lbfgs() -> None:
 
     lbfgs_opt.optimize_controls()
     assert lbfgs_opt.current_best_goal < 0.01
+
+
+@pytest.mark.optimizers
+@pytest.mark.slow
+@pytest.mark.integration
+def test_optim_lbfgs_grad_free() -> None:
+    lbfgs_grad_free_opt = OptimalControl(
+        dir_path=logdir,
+        fid_func=fidelities.average_infid_set,
+        fid_subspace=["Q1", "Q2"],
+        pmap=pmap,
+        algorithm=algorithms.lbfgs_grad_free,
+        options={"maxfun": 5},
+        run_name="grad_free_lbfgs",
+    )
+    lbfgs_grad_free_opt.set_exp(exp)
+
+    lbfgs_grad_free_opt.optimize_controls()
+    assert lbfgs_grad_free_opt.current_best_goal < 0.01
